# Copyright (c) 2021, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Mnemonic handling functionality.

Includes a BIP39 implementation and Electrum-style seed support.

BIP39 is wordlist independent; it just operates on text.  Wordlists are a GUI /
user-friendliness feature.

It is not possible to derive the mnemonic from the seed, so a wallet must store the
mnemonic a user supplied if it wants to be able to display it later.

BIP39 sucks; don't use it for new development.
'''

__all__ = ('BIP39Mnemonic', 'ElectrumMnemonic', 'Wordlists', )


from math import ceil, log

from hashlib import pbkdf2_hmac
from os import urandom
from unicodedata import normalize

from .hashes import sha256, hmac_sha512
from .misc import int_to_be_bytes, be_bytes_to_int, chunks, data_file_path


class Wordlists:
    '''Validate, read and cache word lists.'''

    cache = {}

    @classmethod
    def _text_to_wordlist(cls, text, expected_count):
        '''Convert text to a normalized bip39 wordlist.  Every line is assumed to be a word unless
        it starts with a '#' character or is empty.  Leading and trailing whitespace in
        lines is not significant, and words are converted to lower case.

        Raises: SyntaxError if any word contains a space, or if check is True and there
        are not 2048 words.
        '''
        text = _normalize_text(text)
        lines = [line.strip() for line in text.split('\n')]
        words = [line.lower() for line in lines if line and not line.startswith('#')]
        if any(' ' in word for word in words):
            raise SyntaxError('some words contain whitespace')
        if expected_count is not None and len(words) != expected_count:
            raise SyntaxError(f'text should contain {expected_count} words')
        return words

    @classmethod
    def _read_wordlist(cls, filename, expected_count):
        with open(data_file_path(filename)) as f:
            text = f.read()
        return cls._text_to_wordlist(text, expected_count)

    @classmethod
    def _cached_wordlist(cls, filename, expected_count):
        result = cls.cache.get(filename)
        if not result:
            result = cls.cache[filename] = cls._read_wordlist(filename, expected_count)
        return result

    @classmethod
    def original_electrum_wordlist(cls):
        return cls._cached_wordlist('electrum_old.txt', 1626)

    @classmethod
    def bip39_wordlist(cls, filename):
        return cls._cached_wordlist(filename, 2048)


class BIP39Mnemonic:
    '''BIP39 mnemonic support.'''

    class BadWords(Exception):
        pass

    # This function is split out for testing purposes
    @classmethod
    def _from_entropy(cls, entropy, wordlist):
        size = len(entropy)
        cs_bits = size // 4
        checksum = be_bytes_to_int(sha256(entropy)) >> (256 - cs_bits)
        entropy_cs = be_bytes_to_int(entropy) * (1 << cs_bits) + checksum

        m_len = 3 * size // 4
        parts = []
        for _ in range(m_len):
            entropy_cs, part = divmod(entropy_cs, 2048)
            parts.append(part)

        return ' '.join(wordlist[part] for part in reversed(parts))

    @classmethod
    def generate(cls, wordlist, bits=128):
        '''Return a bip39 mnemonic with bits (128, 160, 192, 224 or 256) entropy.  The mnemonic is
        a space-separated list of words of length (bits + bits // 32) // 11.

        The wordlist is assumed normalized and must have 2048 words.

        Raises: ValueError if the wordlist does not contain 2048 words or bits is invalid.
        '''
        if len(wordlist) != 2048:
            raise ValueError(f'wordlist contains {len(wordlist)} words')

        if bits not in {128, 160, 192, 224, 256}:
            raise ValueError(f'invalid number of entropy bits: {bits}')

        entropy = urandom(bits // 8)

        return cls._from_entropy(entropy, wordlist)

    @classmethod
    def normalize(cls, mnemonic):
        '''Return the normalized mnemonic. Leading and trailing whitespace is removed, whitespace
        is collapsed to a single space, and the mnemonic is converted to lower case.
        '''
        return ' '.join(_normalize_text(mnemonic).split()).lower()

    @classmethod
    def to_seed(cls, mnemonic, passphrase):
        '''Return a 512-bit seed generated from the mnemonic.  The validity of the mnemonic is not
        checked.
        '''
        mnemonic = cls.normalize(mnemonic).encode()
        passphrase = _normalize_text(passphrase).encode()
        return pbkdf2_hmac('sha512', mnemonic, b'mnemonic' + passphrase, iterations=2048)

    @classmethod
    def is_valid(cls, mnemonic, wordlist):
        '''Return true if the mnemonic is valid, i.e., has a correct number of words and that the
        checksum is good.  Wordlist order is significant, and the wordlist is assumed
        normalized.

        Raises ValueError if the wordlist does not contain 2048 words.  Raises
        BIP39.BadWords if the mnemonic contains words not in the wordlist.  The first
        exception argument is a list of the bad words.
        '''
        if len(wordlist) != 2048:
            raise ValueError('wordlist must contain 2048 words')
        words = cls.normalize(mnemonic).split()

        m_len = len(words)
        if m_len not in {12, 15, 18, 21, 24}:
            return False

        def safe_index(word):
            try:
                return wordlist.index(word)
            except ValueError:
                return None

        parts = [safe_index(word) for word in words]
        bad_words = [word for part, word in zip(parts, words) if part is None]
        if bad_words:
            raise cls.BadWords(bad_words)

        entropy = 0
        mult = 1
        for part in reversed(parts):
            entropy += part * mult
            mult *= 2048

        ent_bytes = 4 * m_len // 3
        cs_bits = m_len // 3
        cs_mod = 1 << cs_bits

        checksum = entropy % cs_mod
        entropy = int_to_be_bytes(entropy // cs_mod, size=ent_bytes)

        expected_checksum = be_bytes_to_int(sha256(entropy)) >> (256 - cs_bits)

        return expected_checksum == checksum


class ElectrumMnemonic:
    '''Electrum mnemonic support.'''

    @classmethod
    def generate_new(cls, wordlist, *, prefix='01', bits=132, skip_old=True):
        '''Return a new electrum mnemonic from the wordlist with the given bits of entropy.

        The wordlist can be of arbitrary length and have arbitrary words.
        '''
        if bits < 132:
            raise ValueError(f'insufficient bits to generate a mnemonic: {bits}')

        wordlist_size = len(wordlist)
        if wordlist_size < 2048:
            raise ValueError('wordlist is too short')

        word_count = ceil(bits / log(wordlist_size, 2))
        entropy, delta = _random_integer(word_count, wordlist_size)

        while True:
            mnemonic = cls._from_entropy(entropy, wordlist)
            if cls.is_valid_new(mnemonic, prefix):
                # There is a chance of about 1 in 35,000 it is also a valid old seed
                if not (skip_old and cls.is_valid_old(mnemonic)):
                    return mnemonic
            entropy += delta

    @classmethod
    def generate_old_seed(cls, *, word_count=12):
        '''Return an old electrum seed.  The result is a hex string which is how Electrum used to
        store the seed.

        When converted to a user-facing mnemonic with hex_seed_to_old() it will have the given
        number of words (12 or 24).
        '''
        if word_count not in {12, 24}:
            raise ValueError('word count must be 12 or 24')
        return urandom(word_count * 4 // 3).hex()

    @classmethod
    def is_valid_new(cls, mnemonic, prefix):
        '''Return True if mnemonic is a valid new-style mnemonic for the given prefix.'''
        mnemonic = _normalize_text(mnemonic).lower()
        hmac = hmac_sha512(b'Seed version', mnemonic.encode())
        return hmac.hex().startswith(prefix)

    @classmethod
    def normalize_new(cls, mnemonic):
        return ' '.join(mnemonic.lower().split())

    @classmethod
    def normalize_old(cls, mnemonic):
        return ' '.join(mnemonic.lower().split())

    @classmethod
    def is_valid_old(cls, mnemonic):
        '''Return True if mnemonic is a valid old-style mnemonic.'''
        try:
            cls.old_to_hex_seed(mnemonic)
            return True
        except ValueError:
            pass

        # Check if it's hex of length 32 or 64.
        try:
            return len(bytes.fromhex(mnemonic)) in {16, 32}
        except ValueError:
            return False

    @classmethod
    def old_to_hex_seed(cls, mnemonic):
        mnemonic = cls.normalize_old(mnemonic)
        words = mnemonic.split()
        if len(words) not in {12, 24}:
            raise ValueError('invalid length mnemonic')
        wordlist = Wordlists.original_electrum_wordlist()
        indices = [wordlist.index(word) for word in words]

        def parts():
            n = len(wordlist)
            for pos in range(0, len(words), 3):
                i1, i2, i3 = indices[pos: pos+3]
                part = i1 + n * (((i2 - i1) % n) + n * ((i3 - i2) % n))
                if part >= 4294967296:
                    raise ValueError('mnemonic is not valid')
                yield part

        return ''.join(f'{part:08x}' for part in parts())

    @classmethod
    def hex_seed_to_old(cls, hex_seed):
        if len(hex_seed) not in {32, 64}:
            raise ValueError('hex seed has invalid length')

        def indices(hex_seed, n):
            for chunk in chunks(hex_seed, 8):
                value = int(chunk, 16)
                value, i1 = divmod(value, n)
                i3, i2 = divmod(value, n)
                i2 = (i2 + i1) % n
                i3 = (i3 + i2) % n
                yield i1
                yield i2
                yield i3

        wordlist = Wordlists.original_electrum_wordlist()
        return ' '.join(wordlist[index] for index in indices(hex_seed, len(wordlist)))

    @classmethod
    def _from_entropy(cls, entropy, wordlist):
        '''Covert entropy to a mnemonic.

        Note: this repeats until entropy is exhausted; it is not fixed-length.  The first
        word represents the least significant bits of the entropy, so this is a
        little-endian encoding.
        '''
        def words(entropy, wordlist):
            count = len(wordlist)
            while entropy:
                entropy, rem = divmod(entropy, count)
                yield wordlist[rem]

        return ' '.join(words(entropy, wordlist))


def _normalize_text(text):
    '''Return the normalized text. No elimination of leading or trailing whitespace is
    done, nor is whitespace collapsed.
    '''
    return normalize('NFKD', text)


def _random_integer(word_count, wordlist_size):
    '''Return a random integer sufficient to encode word_count words from a wordlist
    of size wordlist_size.'''

    bits = ceil(word_count * log(wordlist_size, 2)) + 16
    size = (bits + 7) // 8
    # Because of the unfortunate encoding Electrum uses, the most sigificant word cannot have
    # index 0.
    modulus = wordlist_size ** word_count
    min_value = modulus // wordlist_size + 1

    while True:
        value = be_bytes_to_int(urandom(size)) % modulus
        if value <= min_value:
            continue
        delta = 1 if value < (min_value + modulus) // 2 else -1
        return value, delta



# Copyright (c) 2019, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Bitcoin addresses and script classifications.'''

__all__ = (
    'Address', 'P2PKH_Address', 'P2SH_Address',
    'P2PK_Output', 'P2MultiSig_Output', 'OP_RETURN_Output', 'Unknown_Output',
)

from abc import ABC, abstractmethod

from bitcoinx import cashaddr
from .hashes import hash160 as calc_hash160
from .base58 import base58_decode_check, base58_encode_check
from .coin import Bitcoin
from .packing import pack_byte
from .script import Script, Ops, push_item, push_int, item_to_int


class Address(ABC):

    def __init__(self, coin):
        self._coin = coin

    @abstractmethod
    def to_string(self):
        pass

    @classmethod
    def from_string(cls, text, coin):
        '''Construct from an address string.'''
        if len(text) > 35:
            try:
                return cls._from_cashaddr_string(text, coin)
            except ValueError:
                pass

        raw = base58_decode_check(text)

        if len(raw) != 21:
            raise ValueError(f'invalid address: {text}')

        verbyte, hash160 = raw[0], raw[1:]
        if verbyte == coin.P2PKH_verbyte:
            return P2PKH_Address(hash160, coin)
        if verbyte == coin.P2SH_verbyte:
            return P2SH_Address(hash160, coin)

        raise ValueError(f'unknown version byte {verbyte} for coin {coin.name}')

    @classmethod
    def _from_cashaddr_string(cls, text, coin):
        '''Construct from a cashaddress string.'''
        coin = coin or Bitcoin
        prefix = coin.cashaddr_prefix
        if text.upper() == text:
            prefix = prefix.upper()
        if not text.startswith(prefix + ':'):
            text = ':'.join((prefix, text))
        addr_prefix, kind, hash160 = cashaddr.decode(text)
        assert prefix == addr_prefix

        if kind == cashaddr.PUBKEY_TYPE:
            return P2PKH_Address(hash160, coin)
        return P2SH_Address(hash160, coin)

    def coin(self):
        return self._coin

    def __str__(self):
        return self.to_string()


class P2PKH_Address(Address):

    KIND = 'pubkeyhash'

    def __init__(self, hash160, coin):
        super().__init__(coin)
        self._hash160 = _validate_hash160(hash160)

    def __eq__(self, other):
        return isinstance(other, P2PKH_Address) and self._hash160 == other._hash160

    def __hash__(self):
        return hash(self._hash160) + 2

    def to_string(self):
        return base58_encode_check(pack_byte(self._coin.P2PKH_verbyte) + self._hash160)

    def hash160(self):
        return self._hash160

    def to_script_bytes(self):
        return b''.join((
            bytes([Ops.OP_DUP, Ops.OP_HASH160]),
            push_item(self._hash160),
            bytes([Ops.OP_EQUALVERIFY, Ops.OP_CHECKSIG]),
        ))

    def to_script(self):
        return Script(self.to_script_bytes())


class P2SH_Address(Address):

    def __init__(self, hash160, coin):
        super().__init__(coin)
        self._hash160 = _validate_hash160(hash160)

    def __eq__(self, other):
        return isinstance(other, P2SH_Address) and self._hash160 == other._hash160

    def __hash__(self):
        return hash(self._hash160) + 3

    def to_string(self):
        return base58_encode_check(pack_byte(self._coin.P2SH_verbyte) + self._hash160)

    def hash160(self):
        return self._hash160

    def to_script_bytes(self):
        return b''.join((pack_byte(Ops.OP_HASH160), push_item(self._hash160),
                         pack_byte(Ops.OP_EQUAL)))

    def to_script(self):
        return Script(self.to_script_bytes())


class P2PK_Output:

    KIND = 'pubkey'

    def __init__(self, public_key, coin):
        self.public_key = _to_public_key(public_key)
        self._coin = coin

    def __eq__(self, other):
        return isinstance(other, P2PK_Output) and self.public_key == other.public_key

    def __hash__(self):
        return hash(self.public_key) + 1

    def hash160(self):
        return self.public_key.hash160()

    def to_address(self):
        return self.public_key.to_address(coin=self._coin)

    def to_script_bytes(self):
        return push_item(self.public_key.to_bytes()) + pack_byte(Ops.OP_CHECKSIG)

    def to_script(self):
        return Script(self.to_script_bytes())


class P2MultiSig_Output:

    KIND = 'multisig'

    def __init__(self, public_keys, threshold):
        '''public_keys is an iterable.'''
        self.public_keys = tuple(_to_public_key(public_key) for public_key in public_keys)
        self.threshold = threshold
        n = len(self.public_keys)
        if not 1 <= threshold <= n:
            raise ValueError(f'threshold {threshold} is invalid with {n} public keys')

    def __eq__(self, other):
        return (isinstance(other, P2MultiSig_Output)
                and self.public_keys == other.public_keys
                and self.threshold == other.threshold)

    def __hash__(self):
        return hash(self.public_keys) + self.threshold

    def to_script_bytes(self):
        parts = [push_int(self.threshold)]
        parts.extend(push_item(public_key.to_bytes()) for public_key in self.public_keys)
        parts.append(push_int(len(self.public_keys)))
        parts.append(pack_byte(Ops.OP_CHECKMULTISIG))
        return b''.join(parts)

    def to_script(self):
        return Script(self.to_script_bytes())

    def hash160(self):
        return calc_hash160(self.to_script_bytes())

    def public_key_count(self):
        return len(self.public_keys)

    @classmethod
    def from_template(cls, *items):
        threshold, *public_keys, count = items
        n = len(public_keys)
        count = item_to_int(count)
        if count != n:
            raise ValueError(f'received {n} public keys but {count} as their count')
        return cls(public_keys, item_to_int(threshold))


class OP_RETURN_Output:
    '''This class indicates the script is an OP_RETURN script.'''

    KIND = 'op_return'

    def __eq__(self, other):
        return isinstance(other, OP_RETURN_Output)

    def __hash__(self):
        return 28

    def to_script_bytes(self):
        return pack_byte(Ops.OP_RETURN)

    def to_script(self):
        return Script(self.to_script_bytes())

    @classmethod
    def from_template(cls, *_items):
        return cls()


class Unknown_Output:

    KIND = 'unknown'

    def to_script_bytes(self):
        raise RuntimeError('no canonical script')

    def to_script(self):
        return Script(self.to_script_bytes())


def _validate_hash160(hash160):
    if not isinstance(hash160, bytes):
        raise TypeError('hash160 must be bytes')
    if len(hash160) != 20:
        raise ValueError('hash160 must be 20 bytes')
    return hash160


def _to_public_key(obj):
    '''Convert obj a PublicKey object.'''
    from .keys import PublicKey
    if isinstance(obj, PublicKey):
        return obj
    if isinstance(obj, str):
        return PublicKey.from_hex(obj)
    return PublicKey.from_bytes(obj)


# def _to_signature(obj):
#     '''Convert obj a Signature object.'''
#     if isinstance(obj, Signature):
#         return obj
#     if isinstance(obj, str):
#         return Signature.from_hex(obj)
#     return Signature(obj)



# Copyright (c) 2021, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


__all__ = (
    'JSONFlags',
)


from enum import IntFlag


class JSONFlags(IntFlag):
    '''Flags controlling conversion of transactions and scripts to JSON.'''
    # Include the index of each input
    ENUMERATE_INPUTS = 1 << 0
    # Include the index of each output
    ENUMERATE_OUTPUTS = 1 << 1
    # Include the transaction size in bytes
    SIZE = 1 << 2
    # Include a human-readable description of the locktime constraint is output
    LOCKTIME_MEANING = 1 << 3
    # Include classification of output scripts
    CLASSIFY_OUTPUT_SCRIPT = 1 << 4
    # Display signature sighashes as text
    SIGHASH_MEANING = 1 << 5



# Copyright (c) 2019-2021, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Signatures of various kinds.'''

__all__ = (
    'Signature', 'SigHash',
    'der_signature_to_compact', 'compact_signature_to_der', 'InvalidSignatureError',
)

from base64 import b64decode
from binascii import Error as binascii_Error

from electrumsv_secp256k1 import ffi, lib

from .misc import CONTEXT, be_bytes_to_int
from .packing import pack_byte


CDATA_SIG_LENGTH = 64
# This is for the ECDSA signature without the sighash suffix
MAX_SIG_LENGTH = 72
MISSING_SIG_BYTES = b'\xff'


class InvalidSignatureError(ValueError):
    pass


def _cdata_recsig(recoverable_sig):
    if len(recoverable_sig) != 65:
        raise InvalidSignatureError('invalid recoverable signature')
    cdata_recsig = ffi.new('secp256k1_ecdsa_recoverable_signature *')
    recid = recoverable_sig[-1]
    if not 0 <= recid <= 3:
        raise InvalidSignatureError('invalid recoverable signature')
    if not lib.secp256k1_ecdsa_recoverable_signature_parse_compact(
            CONTEXT, cdata_recsig, recoverable_sig, recid):
        raise InvalidSignatureError('invalid recoverable signature')
    return cdata_recsig


def public_key_from_recoverable_signature(recoverable_sig, msg_hash):
    cdata_recsig = _cdata_recsig(recoverable_sig)
    public_key = ffi.new('secp256k1_pubkey *')
    if not lib.secp256k1_ecdsa_recover(CONTEXT, public_key, cdata_recsig, msg_hash):
        raise InvalidSignatureError('invalid recoverable signature')
    return public_key


def _cdata_signature_to_der(signature):
    '''Return a DER-serialized signature; bytes of length at most 72.'''
    size = ffi.new('size_t *', MAX_SIG_LENGTH)
    data = ffi.new(f'unsigned char [{MAX_SIG_LENGTH}]')
    result = lib.secp256k1_ecdsa_signature_serialize_der(CONTEXT, data, size, signature)
    # Failure should never happen - MAX_SIG_LENGTH bytes is always enough
    assert result
    return bytes(ffi.buffer(data, size[0]))


def _der_signature_to_cdata(der_sig):
    cdata_sig = ffi.new('secp256k1_ecdsa_signature *')
    if not lib.secp256k1_ecdsa_signature_parse_der(CONTEXT, cdata_sig, der_sig, len(der_sig)):
        raise InvalidSignatureError('invalid DER-encoded signature')
    return cdata_sig


def der_signature_to_compact(der_sig):
    '''Returns 64 bytes representing r and s as concatenated 32-byte big-endian numbers.'''
    cdata_sig = _der_signature_to_cdata(der_sig)
    compact_sig = ffi.new('unsigned char [64]')
    lib.secp256k1_ecdsa_signature_serialize_compact(CONTEXT, compact_sig, cdata_sig)
    return bytes(ffi.buffer(compact_sig))


def compact_signature_to_der(compact_sig):
    if not (isinstance(compact_sig, bytes) and len(compact_sig) == 64):
        raise InvalidSignatureError('compact signature must be 64 bytes')
    cdata_sig = ffi.new('secp256k1_ecdsa_signature *')
    lib.secp256k1_ecdsa_signature_parse_compact(CONTEXT, cdata_sig, compact_sig)
    return _cdata_signature_to_der(cdata_sig)


def sign_der(msg_hash, secret):
    cdata_sig = ffi.new('secp256k1_ecdsa_signature *')
    if not lib.secp256k1_ecdsa_sign(CONTEXT, cdata_sig, msg_hash, secret, ffi.NULL, ffi.NULL):
        raise ValueError('invalid private key')
    return _cdata_signature_to_der(cdata_sig)


def sign_recoverable(msg_hash, secret):
    '''Sign a message hash and return a 65-byte recoverable signature.  This is a 64-byte
    compact signature with a recovery ID byte appended; and from which the public key can
    be immediately recovered.
    '''
    rec_signature = ffi.new('secp256k1_ecdsa_recoverable_signature *')
    if not lib.secp256k1_ecdsa_sign_recoverable(CONTEXT, rec_signature, msg_hash,
                                                secret, ffi.NULL, ffi.NULL):
        raise ValueError('invalid private key')

    # Serialize its as a 65-byte compact recoverable signature
    output = ffi.new(f'unsigned char [{CDATA_SIG_LENGTH}]')
    recid = ffi.new('int *')

    lib.secp256k1_ecdsa_recoverable_signature_serialize_compact(
        CONTEXT, output, recid, rec_signature)

    # recid is 0, 1, 2 or 3.
    return bytes(ffi.buffer(output, CDATA_SIG_LENGTH)) + bytes([recid[0]])


def verify_recoverable_signature(recoverable_sig, msg_hash, public_key):
    # Convert a 65-byte recoverable sig to a CDATA one
    cdata_sig = ffi.new('secp256k1_ecdsa_signature *')
    cdata_recsig = _cdata_recsig(recoverable_sig)
    # Always succeeds
    lib.secp256k1_ecdsa_recoverable_signature_convert(CONTEXT, cdata_sig, cdata_recsig)
    return bool(lib.secp256k1_ecdsa_verify(CONTEXT, cdata_sig, msg_hash, public_key))


def verify_der_signature(der_sig, msg_hash, public_key):
    '''Verify a der-encoded signature.  Return True if good otherwise False.'''
    cdata_sig = _der_signature_to_cdata(der_sig)
    return bool(lib.secp256k1_ecdsa_verify(CONTEXT, cdata_sig, msg_hash, public_key))


def to_recoverable_signature(message_sig):
    '''Return a recoverable signature from a message signature.'''
    if isinstance(message_sig, str):
        try:
            message_sig = b64decode(message_sig, validate=True)
        except binascii_Error:
            raise InvalidSignatureError('invalid base64 encoding of message signature') from None
    if not isinstance(message_sig, bytes) or len(message_sig) != 65:
        raise InvalidSignatureError('message signature must be 65 bytes')
    if not 27 <= message_sig[0] < 35:
        raise InvalidSignatureError('invalid message signature format')
    return message_sig[1:] + pack_byte((message_sig[0] - 27) & 3)


def to_message_signature(recoverable_sig, compressed):
    leading_byte = 27 + (4 if compressed else 0) + recoverable_sig[-1]
    return pack_byte(leading_byte) + recoverable_sig[:64]


class SigHash(int):

    @property
    def base(self):
        return SigHash(self & 0x1f)

    @property
    def anyone_can_pay(self):
        '''Returns True if ANYONE_CAN_PAY is set.'''
        return bool(self & SigHash.ANYONE_CAN_PAY)

    def to_string(self):
        kinds = []
        if self & 3:
            kinds.append(('ALL', 'NONE', 'SINGLE')[(self & 3) - 1])
        if self & self.ANYONE_CAN_PAY:
            kinds.append('ANYONE_CAN_PAY')
        if self & self.FORKID:
            kinds.append('FORKID')
        residual = self & ~(self.ANYONE_CAN_PAY | self.FORKID | 3)
        if residual:
            kinds.append(f'0x{residual:02x}')
        return '|'.join(kinds)


class Signature:
    '''A bitcoin DER signature, as raw bytes.'''

    def __init__(self, raw):
        '''Raw is a der-encoded signature plus a single sighash byte, or MISSING_SIG_BYTES.

        Raises InvalidSignatureError.'''
        if raw != MISSING_SIG_BYTES:
            # Validate the DER encoding
            der_signature_to_compact(raw[:-1])
        self._raw = raw

    def __eq__(self, other):
        '''A signature equals anything buffer-like with the same bytes representation.'''
        return (isinstance(other, (bytes, bytearray, memoryview))
                or hasattr(other, '__bytes__')) and self._raw == bytes(other)

    def __hash__(self):
        return hash(self._raw)

    @classmethod
    def from_hex(cls, hex_str):
        '''Instantiate from a hexadecimal string.'''
        return cls(bytes.fromhex(hex_str))

    def to_hex(self):
        '''Return the script signature as a hexadecimal string.'''
        return self._raw.hex()

    @classmethod
    def from_der_sig(cls, der_sig, sighash):
        return cls(der_sig + pack_byte(sighash))

    def __bytes__(self):
        return self._raw

    def to_bytes(self):
        return self._raw

    def is_present(self):
        return self._raw != MISSING_SIG_BYTES

    def to_compact(self):
        '''The 32-byte r and s values concatenated.'''
        return der_signature_to_compact(self.der_signature)

    def r_value(self):
        '''The r value as an integer.'''
        return be_bytes_to_int(self.to_compact()[:32])

    def s_value(self):
        '''The s value as an integer.'''
        return be_bytes_to_int(self.to_compact()[32:])

    def to_string(self):
        '''The signature as an ASM string.'''
        return self._raw[:-1].hex() + '[' + SigHash(self._raw[-1]).to_string() + ']'

    @property
    def der_signature(self):
        if self._raw == MISSING_SIG_BYTES:
            raise InvalidSignatureError('signature is missing')
        return self._raw[:-1]

    @property
    def sighash(self):
        if self._raw == MISSING_SIG_BYTES:
            raise InvalidSignatureError('signature is missing')
        return SigHash(self._raw[-1])


# Sighash values
SigHash.ALL = SigHash(0x01)
SigHash.NONE = SigHash(0x02)
SigHash.SINGLE = SigHash(0x03)
SigHash.FORKID = SigHash(0x40)
SigHash.ANYONE_CAN_PAY = SigHash(0x80)

Signature.MISSING = Signature(MISSING_SIG_BYTES)



# Copyright (c) 2019-2021, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


__all__ = (
    'Tx', 'TxInput', 'TxOutput',
)

import attr
import datetime
from io import BytesIO

from .consts import JSONFlags
from .hashes import hash_to_hex_str, double_sha256
from .packing import (
    pack_le_int32, pack_le_uint32, pack_varbytes, pack_le_int64, pack_list, varint_len,
    read_le_int32, read_le_uint32, read_varbytes, read_le_int64, read_list,
)
from .script import Script
from .signature import SigHash


ZERO = bytes(32)
UINT32_MAX = (1 << 32) - 1
LOCKTIME_THRESHOLD = 500_000_000


@attr.s(slots=True)
class Tx:
    '''A bitcoin transaction.'''
    version = attr.ib()
    inputs = attr.ib()
    outputs = attr.ib()
    locktime = attr.ib()

    def is_coinbase(self):
        '''Return True iff the tx is a coinbase transaction.'''
        return self.inputs[0].is_coinbase()

    @classmethod
    def read(cls, read):
        return cls(
            read_le_int32(read),
            read_list(read, TxInput.read),
            read_list(read, TxOutput.read),
            read_le_uint32(read),
        )

    @classmethod
    def from_bytes(cls, raw):
        return cls.read(BytesIO(raw).read)

    def to_bytes(self):
        return b''.join((
            pack_le_int32(self.version),
            pack_list(self.inputs, TxInput.to_bytes),
            pack_list(self.outputs, TxOutput.to_bytes),
            pack_le_uint32(self.locktime),
        ))

    @classmethod
    def from_hex(cls, hex_str):
        return cls.from_bytes(bytes.fromhex(hex_str))

    def to_hex(self):
        return self.to_bytes().hex()

    def _hash_prevouts(self):
        preimage = b''.join(txin.prevout_bytes() for txin in self.inputs)
        return double_sha256(preimage)

    def _hash_sequence(self):
        preimage = b''.join(pack_le_uint32(txin.sequence) for txin in self.inputs)
        return double_sha256(preimage)

    def _hash_outputs(self):
        preimage = b''.join(txout.to_bytes() for txout in self.outputs)
        return double_sha256(preimage)

    def signature_hash(self, input_index, value, script, *, sighash=None):
        if not 0 <= input_index < len(self.inputs):
            raise IndexError(f'invalid input index: {input_index}')
        if value < 0:
            raise ValueError(f'value cannot be negative: {value}')
        if sighash is None:
            sighash = SigHash(SigHash.ALL | SigHash.FORKID)
        if not isinstance(sighash, SigHash):
            raise TypeError('sighash must be a SigHash instance')

        txin = self.inputs[input_index]
        hash_prevouts = hash_sequence = hash_outputs = ZERO

        sighash_not_single_none = sighash.base not in (SigHash.SINGLE, SigHash.NONE)
        if not sighash.anyone_can_pay:
            hash_prevouts = self._hash_prevouts()
            if sighash_not_single_none:
                hash_sequence = self._hash_sequence()
        if sighash_not_single_none:
            hash_outputs = self._hash_outputs()
        elif (sighash.base == SigHash.SINGLE and input_index < len(self.outputs)):
            hash_outputs = double_sha256(self.outputs[input_index].to_bytes())

        preimage = b''.join((
            pack_le_int32(self.version),
            hash_prevouts,
            hash_sequence,
            txin.to_bytes_for_signature(value, script),
            hash_outputs,
            pack_le_uint32(self.locktime),
            pack_le_uint32(sighash),
        ))

        return double_sha256(preimage)

    def are_inputs_final(self):
        '''Return True if all inputs are final.'''
        return all(txin.is_final() for txin in self.inputs)

    def is_final_for_block(self, block_height, timestamp):
        '''Return True if a transaction is final for the given height and timestamp.

        Transactions cannot be mined unless final.  BIP113 introduced the consensus rule
        that the timestamp is the MTP of the previous block; the median timestamp of the
        11 blocks with heights [block_height-11, block_height-1].  The previous consensus
        rule was to use the block's timestamp.
        '''
        return (
            self.locktime == 0 or
            self.locktime < (block_height if self.locktime < LOCKTIME_THRESHOLD else timestamp) or
            self.are_inputs_final()
        )

    def hash(self):
        '''Return the transaction hash.   Only makes sense for fully-signed transactions.'''
        return double_sha256(self.to_bytes())

    def hex_hash(self):
        '''Return the transaction hash as a hex string if it is complete, otherwise None.'''
        tx_hash = self.hash()
        return hash_to_hex_str(tx_hash) if tx_hash else None

    @staticmethod
    def size_io(inputs, outputs):
        '''Return the size of a transaction with the given inputs and outputs.

        Faster than serializing and taking its length.
        '''
        # 8 for version and locktime
        return (varint_len(len(inputs)) + sum(tx_input.size() for tx_input in inputs) +
                varint_len(len((outputs))) + sum(output.size() for output in outputs) + 8)

    def size(self):
        '''Return the size of the transaction.  More efficient than serializing to bytes.'''
        return self.size_io(self.inputs, self.outputs)

    def total_output_value(self):
        '''Return the sum of the output values.'''
        return sum(output.value for output in self.outputs)

    def to_json(self, flags, coin):
        result = {
            'version': self.version,
            'nInputs': len(self.inputs),
            'vin': [input.to_json(flags, index) for index, input in enumerate(self.inputs)],
            'nOutputs': len(self.outputs),
            'vout': [output.to_json(flags, coin, index)
                     for index, output in enumerate(self.outputs)],
            'locktime': self.locktime,
            'hash': self.hex_hash(),
        }
        if flags & JSONFlags.SIZE:
            result['size'] = self.size()
        if flags & JSONFlags.LOCKTIME_MEANING:
            result['locktimeMeaning'] = locktime_description(self.locktime)
        return result


@attr.s(slots=True, repr=False)
class TxInput:
    '''A bitcoin transaction input.'''
    prev_hash = attr.ib()
    prev_idx = attr.ib()
    script_sig = attr.ib()
    sequence = attr.ib()

    def is_coinbase(self):
        '''Return True iff the input is the single input of a coinbase transaction.'''
        return self.prev_idx == UINT32_MAX and self.prev_hash == ZERO

    @classmethod
    def read(cls, read):
        return cls(
            read(32),                       # prev_hash
            read_le_uint32(read),           # prev_idx
            Script(read_varbytes(read)),    # script_sig
            read_le_uint32(read),           # sequence
        )

    def prevout_bytes(self):
        return self.prev_hash + pack_le_uint32(self.prev_idx)

    def to_bytes_for_signature(self, value, script):
        return b''.join((
            self.prevout_bytes(),
            pack_varbytes(bytes(script)),
            pack_le_int64(value),
            pack_le_uint32(self.sequence),
        ))

    def to_bytes(self):
        return b''.join((
            self.prevout_bytes(),
            pack_varbytes(bytes(self.script_sig)),
            pack_le_uint32(self.sequence),
        ))

    def to_hex(self):
        return self.to_bytes().hex()

    @classmethod
    def from_bytes(cls, raw):
        return cls.read(BytesIO(raw).read)

    @classmethod
    def from_hex(cls, hex_str):
        return cls.from_bytes(bytes.fromhex(hex_str))

    def size(self):
        '''Return the serialized size of the input in bytes.'''
        n = len(self.script_sig)
        # 40 for prevout and sequence
        return 40 + varint_len(n) + n

    def is_final(self):
        return self.sequence == 0xffffffff

    def to_json(self, flags, index=None):
        if self.is_coinbase():
            result = {
                'coinbase': self.script_sig.to_hex(),
                'text': self.script_sig.to_bytes().decode(errors='replace'),
                'sequence': self.sequence,
            }
        else:
            result = {
                'hash': self.prev_hash.hex(),
                'idx': self.prev_idx,
                'script': self.script_sig.to_json(flags, True, None),
                'sequence': self.sequence,
            }

        if flags & JSONFlags.ENUMERATE_INPUTS and index is not None:
            result['nInput'] = index
        return result

    def __repr__(self):
        return (
            f'TxInput(prev_hash="{hash_to_hex_str(self.prev_hash)}", prev_idx={self.prev_idx}, '
            f'script_sig="{self.script_sig}", sequence={self.sequence})'
        )


@attr.s(slots=True, repr=False)
class TxOutput:
    '''A bitcoin transaction output.'''
    value = attr.ib()
    script_pubkey = attr.ib()

    @classmethod
    def read(cls, read):
        return cls(
            read_le_int64(read),           # value
            Script(read_varbytes(read)),   # script_pubkey
        )

    def to_bytes(self):
        return b''.join((
            pack_le_int64(self.value),
            pack_varbytes(bytes(self.script_pubkey)),
        ))

    def size(self):
        '''Return the serialized size of the output in bytes.'''
        n = len(self.script_pubkey)
        # 8 for the value
        return 8 + varint_len(n) + n

    def to_hex(self):
        return self.to_bytes().hex()

    @classmethod
    def from_bytes(cls, raw):
        return cls.read(BytesIO(raw).read)

    @classmethod
    def from_hex(cls, hex_str):
        return cls.from_bytes(bytes.fromhex(hex_str))

    def to_json(self, flags, coin, index=None):
        result = {
            'value': self.value,
            'script': self.script_pubkey.to_json(flags, False, coin),
        }
        if flags & JSONFlags.ENUMERATE_OUTPUTS and index is not None:
            result['nOutput'] = index
        return result

    def __repr__(self):
        return (
            f'TxOutput(value={self.value}, script_pubkey="{self.script_pubkey}")'
        )


def locktime_description(locktime):
    '''A human-readable description of meaning of locktime.'''
    if locktime == 0:
        return 'valid in any block'
    if locktime < 500_000_000:
        return f'valid in blocks with height greater than {locktime:,d}'
    utc = datetime.datetime.utcfromtimestamp(locktime)
    return f'valid in blocks with MTP greater than {utc.isoformat(" ")} UTC'
